# -*- coding: utf-8 -*-
"""
Created on Sun Jun  6 15:06:41 2021

@author: Sashka
"""
from pickle import NONE
import torch
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
from TEDEouS.input_preprocessing import grid_prepare, bnd_prepare, operator_prepare
from TEDEouS.metrics import point_sort_shift_loss,point_sort_shift_loss_batch
from TEDEouS.input_preprocessing import bnd_prepare_matrix,operator_prepare_matrix
from TEDEouS.input_preprocessing import operator_prepare_autograd,bnd_prepare_autograd
from TEDEouS.input_preprocessing import op_dict_to_list
from TEDEouS.metrics import matrix_loss,autograd_loss
from TEDEouS.cache import cache_lookup,cache_retrain,save_model,cache_lookup_autograd







def solution_print(prepared_grid,model,title=None):
    if prepared_grid.shape[1] == 2:
        nvars_model = model(prepared_grid).shape[1]
        #plt.ion()
        fig1 = plt.figure()
        ax1 = fig1.add_subplot(projection='3d')

        if title!=None:
            ax1.set_title(title)
            # ax2.set_title(title)
        if nvars_model == 1:
            ax1.plot_trisurf(prepared_grid[:, 0].detach().numpy().reshape(-1), prepared_grid[:, 1].detach().numpy().reshape(-1),
                        model(prepared_grid).detach().numpy().reshape(-1), cmap=cm.jet, linewidth=0.2, alpha=1)
            ax1.set_xlabel("x1")
            ax1.set_ylabel("x2")

        elif nvars_model == 2:
            fig2 = plt.figure()
            ax2 = fig2.add_subplot(projection='3d')
            ax1.plot_trisurf(prepared_grid[:, 0].reshape(-1), prepared_grid[:, 1].reshape(-1),
                        model(prepared_grid)[:,0].detach().numpy().reshape(-1), cmap=cm.jet, linewidth=0.2, alpha=1)
            ax2.plot_trisurf(prepared_grid[:, 0].reshape(-1), prepared_grid[:, 1].reshape(-1),
                        model(prepared_grid)[:,1].detach().numpy().reshape(-1), cmap=cm.jet, linewidth=0.2, alpha=1)
            ax2.set_xlabel("x1")
            ax2.set_ylabel("x2")
        
        
        #plt.show(block=False)
        plt.show()
        #plt.pause(0.001)

    if prepared_grid.shape[1] == 1:
        fig = plt.figure()
        plt.scatter(prepared_grid.detach().numpy().reshape(-1), model(prepared_grid).detach().numpy().reshape(-1))
        #plt.show(block=False)
        plt.show()


def solution_print_mat(grid,model,title=None):
    if grid.shape[0] == 1:
        fig = plt.figure()
        plt.scatter(grid.reshape(-1), model.detach().numpy().reshape(-1))
        plt.show()
    if grid.shape[0] == 2:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        if title!=None:
            ax.set_title(title)
        ax.plot_trisurf(grid[0].reshape(-1), grid[1].reshape(-1),
                        model.reshape(-1).detach().numpy(), cmap=cm.jet, linewidth=0.2, alpha=1)
        ax.set_xlabel("x1")
        ax.set_ylabel("x2")
        plt.show()



def create_random_fn(eps):
    def randomize_params(m):
      if type(m)==torch.nn.Linear or type(m)==torch.nn.Conv2d:
        m.weight.data=m.weight.data+(2*torch.randn(m.weight.size())-1)*eps#Random weight initialisation
        m.bias.data=m.bias.data+(2*torch.randn(m.bias.size())-1)*eps
    return randomize_params




def point_sort_shift_solver(grid, model, operator, bconds, grid_point_subset=['central'], lambda_bound=10,
                            verbose=False, learning_rate=1e-4, eps=1e-5, tmin=1000, tmax=1e5, h=0.001,
                            use_cache=True,cache_dir='../cache/',cache_verbose=False,
                            batch_size=None,save_always=False,lp_par=None,print_every=100,
                            patience=5,loss_oscillation_window=100,no_improvement_patience=1000,
                            model_randomize_parameter=0,optimizer='Adam',print_plot = True):
    # prepare input data to uniform format 
    if model==None:
            model = torch.nn.Sequential(
                torch.nn.Linear(grid.shape[-1], 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 1)
            )
    prepared_grid,grid_dict,point_type = grid_prepare(grid)
    prepared_bconds = bnd_prepare(bconds, prepared_grid,grid_dict, h=h)
    full_prepared_operator = operator_prepare(operator, grid_dict, subset=grid_point_subset, true_grid=grid, h=h)
    
    r=create_random_fn(model_randomize_parameter)   
    #  use cache if needed
    if use_cache:
        cache_checkpoint,min_loss=cache_lookup(prepared_grid, full_prepared_operator, prepared_bconds,cache_dir=cache_dir
                                               ,nmodels=None,verbose=cache_verbose,lambda_bound=lambda_bound,norm=lp_par)
        model, optimizer_state= cache_retrain(model,cache_checkpoint,grid,verbose=cache_verbose)
  
        model.apply(r)


    if optimizer=='Adam':
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    elif optimizer=='SGD':
        optimizer = torch.optim.SGD(model.parameters(), lr=learning_rate)
    elif optimizer=='LBFGS':
        optimizer = torch.optim.LBFGS(model.parameters(), lr=learning_rate)
    else:
        print('Wrong optimizer chosen, optimization was not performed')
        return model

    if not use_cache:
        min_loss = point_sort_shift_loss(model, prepared_grid, full_prepared_operator, prepared_bconds, lambda_bound=lambda_bound)
    
    save_cache=False
    
    if min_loss>0.1 or save_always:
        save_cache=True
    
    
    # standard NN stuff
    if verbose:
        print('-1 {}'.format(min_loss))
    
    t = 0
    
    last_loss=np.zeros(loss_oscillation_window)+float(min_loss)
    line=np.polyfit(range(loss_oscillation_window),last_loss,1)
    

    def closure():
        nonlocal cur_loss
        optimizer.zero_grad()
        if batch_size==None:
            loss = point_sort_shift_loss(model, prepared_grid, full_prepared_operator, prepared_bconds, lambda_bound=lambda_bound,norm=lp_par)
        else:
            loss=point_sort_shift_loss_batch(model, prepared_grid, point_type, operator, bconds,subset=grid_point_subset, lambda_bound=lambda_bound,batch_size=batch_size,h=h,norm=lp_par)
        loss.backward()
        cur_loss = loss.item()
        return loss
    
    stop_dings=0
    t_imp_start=0
    # to stop train proceduce we fit the line in the loss data
    #if line is flat enough 5 times, we stop the procedure
    cur_loss=min_loss
    while stop_dings<=patience:
        optimizer.step(closure)

        last_loss[t%loss_oscillation_window]=cur_loss

        
        if cur_loss<min_loss:
            min_loss=cur_loss
            t_imp_start=t
        if t%loss_oscillation_window==0:
            line=np.polyfit(range(loss_oscillation_window),last_loss,1)
            if abs(line[0]/cur_loss) < eps and t>0:
                stop_dings+=1
                model.apply(r)
                if verbose:
                    print('Oscillation near the same loss')
                    print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
                    solution_print(prepared_grid,model,title='Iteration = ' + str(t))
        

        if (t-t_imp_start==no_improvement_patience) and verbose:
            print('No improvement in '+str(no_improvement_patience)+' steps')
            t_imp_start=t
            stop_dings+=1
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            solution_print(prepared_grid,model,title='Iteration = ' + str(t))
            
        if print_every!=None and (t % print_every == 0) and verbose:
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            solution_print(prepared_grid,model,title='Iteration = ' + str(t))

        t += 1
        if t > tmax:
            break
    if (save_cache and use_cache) or save_always:
        save_model(model,model.state_dict(),optimizer.state_dict(),cache_dir=cache_dir,name=None)
    return model


def nn_optimizer(grid, model, operator, bconds, grid_point_subset=['central'], lambda_bound=10,
                            verbose=False, learning_rate=1e-4, eps=1e-5, tmin=1000, tmax=1e5, h=0.001,
                            use_cache=True,cache_dir='../cache/',cache_verbose=False,
                            batch_size=None,save_always=False,lp_par=None,print_every=100,
                            patience=5,loss_oscillation_window=100,no_improvement_patience=1000,
                            model_randomize_parameter=0,optimizer='Adam'):
    # prepare input data to uniform format 
    
    prepared_grid,grid_dict,point_type = grid_prepare(grid)
    prepared_bconds = bnd_prepare(bconds, prepared_grid,grid_dict, h=h)
    full_prepared_operator = operator_prepare(operator, grid_dict, subset=grid_point_subset, true_grid=grid, h=h)
    
    r=create_random_fn(model_randomize_parameter)   
    #  use cache if needed
    if use_cache:
        cache_checkpoint,min_loss=cache_lookup(prepared_grid, full_prepared_operator, prepared_bconds,cache_dir=cache_dir
                                               ,nmodels=None,verbose=cache_verbose,lambda_bound=lambda_bound,norm=lp_par)
        model, optimizer_state= cache_retrain(model,cache_checkpoint,grid,verbose=cache_verbose)
  
        model.apply(r)

    if optimizer=='Adam':
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    elif optimizer=='SGD':
        optimizer = torch.optim.SGD(model.parameters(), lr=learning_rate)
    elif optimizer=='LBFGS':
        optimizer = torch.optim.LBFGS(model.parameters(), lr=learning_rate)
    else:
        print('Wrong optimizer chosen, optimization was not performed')
        return model
    

    if not use_cache:
        min_loss = point_sort_shift_loss(model, prepared_grid, full_prepared_operator, prepared_bconds, lambda_bound=lambda_bound)
    
    save_cache=False
    
    if min_loss>0.1 or save_always:
        save_cache=True
    
    
    # standard NN stuff
    if verbose:
        print('-1 {}'.format(min_loss))
    
    t = 0
    
    last_loss=np.zeros(loss_oscillation_window)+float(min_loss)
    line=np.polyfit(range(loss_oscillation_window),last_loss,1)
    

    def closure():
        nonlocal cur_loss
        optimizer.zero_grad()
        if batch_size==None:
            loss = point_sort_shift_loss(model, prepared_grid, full_prepared_operator, prepared_bconds, lambda_bound=lambda_bound,norm=lp_par)
        else:
            loss=point_sort_shift_loss_batch(model, prepared_grid, point_type, operator, bconds,subset=grid_point_subset, lambda_bound=lambda_bound,batch_size=batch_size,h=h,norm=lp_par)
        loss.backward()
        cur_loss = loss.item()
        return loss
    
    stop_dings=0
    t_imp_start=0
    # to stop train proceduce we fit the line in the loss data
    #if line is flat enough 5 times, we stop the procedure
    cur_loss=min_loss
    while stop_dings<=patience:
        optimizer.step(closure)

        last_loss[t%loss_oscillation_window]=cur_loss
        
        if cur_loss<min_loss:
            min_loss=cur_loss
            t_imp_start=t
        if t%loss_oscillation_window==0:
            line=np.polyfit(range(loss_oscillation_window),last_loss,1)

            if abs(line[0]/cur_loss) < eps and t>0:
                stop_dings+=1
                model.apply(r)
                if verbose:
                    print('Oscillation near the same loss')
                    print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
                    if print_plot:
                        solution_print(prepared_grid,model,title='Iteration = ' + str(t))

        if (t-t_imp_start==no_improvement_patience) and verbose:
            print('No improvement in '+str(no_improvement_patience)+' steps')
            t_imp_start=t
            stop_dings+=1
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            if print_plot:
                solution_print(prepared_grid,model,title='Iteration = ' + str(t))
            
        if print_every!=None and (t % print_every == 0) and verbose:
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            if print_plot:
                solution_print(prepared_grid,model,title='Iteration = ' + str(t))


        t += 1
        if t > tmax:
            break
    if (save_cache and use_cache) or save_always:
        save_model(model,model.state_dict(),optimizer.state_dict(),cache_dir=cache_dir,name=None)
    return model





def lbfgs_solution(model, grid, operator, norm_lambda, bcond, rtol=1e-6,atol=0.01,nsteps=10000):
    
    unified_operator = operator_prepare_matrix(operator)

    b_prepared = bnd_prepare_matrix(bcond, grid)

    # optimizer = torch.optim.Adam([model.requires_grad_()], lr=1e-4)
    optimizer = torch.optim.LBFGS([model.requires_grad_()], lr=1e-3)
   
    
    def closure():
        nonlocal cur_loss
        optimizer.zero_grad()

        loss = matrix_loss(model, grid, unified_operator, b_prepared, lambda_bound=norm_lambda)
        loss.backward()
        cur_loss = loss.item()

        return loss

    cur_loss = float('inf')
    # tol = 1e-20

    for i in range(nsteps):
        optimizer.zero_grad()
        past_loss = cur_loss
        
        # loss = matrix_loss(model, grid, unified_operator, b_prepared, lambda_bound=norm_lambda)
        optimizer.step(closure)
        
        # optimizer.zero_grad()
        # loss.backward()
        # optimizer.step()
        # cur_loss = loss.item()
        
        
        if i % 1000 == 0:
            print('i={} loss={}'.format(i, cur_loss))

        if abs(cur_loss - past_loss) / abs(cur_loss) < rtol:
        #     # print("sosholsya")
            break
        if abs(cur_loss) < atol:
        #     # print("sosholsya")
            break

    return model.detach()

from copy import deepcopy

def matrix_cache_lookup(grid, model, operator, bconds,h=0.001,model_randomize_parameter=0,cache_dir="../cache",
                            lambda_bound=10):
    # prepare input data to uniform format 
    if type(operator)==dict:
        op_list=op_dict_to_list(operator)
    else:
        op_list=deepcopy(operator)

    for term in op_list:
        if type(term[0])==torch.Tensor:
            term[0]=term[0].reshape(-1)
        if callable(term[0]):
            print("Warning: coefficient is callable, it may lead to wrong cache item choice")

    prepared_grid,grid_dict,point_type = grid_prepare(grid)
    prepared_bconds = bnd_prepare(bconds, prepared_grid,grid_dict, h=h)
    full_prepared_operator = operator_prepare(op_list, grid_dict, true_grid=grid, h=h)
    
    r=create_random_fn(model_randomize_parameter)   
    #  use cache if needed

    cache_checkpoint,min_loss=cache_lookup(prepared_grid, full_prepared_operator, prepared_bconds,cache_dir=cache_dir
                                           ,nmodels=None,verbose=True,lambda_bound=lambda_bound)
    model, optimizer_state= cache_retrain(model,cache_checkpoint,grid,verbose=False)
  
    model.apply(r)
    
    return model



def matrix_optimizer(grid, model, operator, bconds, lambda_bound=10,
                            verbose=False, learning_rate=1e-4, eps=1e-5, tmin=1000, tmax=1e5,
                            use_cache=True,cache_dir='../cache/',cache_verbose=False,
                            batch_size=None,save_always=False,lp_par=None,print_every=100,
                            patience=5,loss_oscillation_window=100,no_improvement_patience=1000,
                            model_randomize_parameter=1e-5,optimizer='LBFGS',cache_model=None, coord_list=None):
    # prepare input data to uniform format 
    
    #if coord_list==None:
     #   print("Error: No coord_list supplied, using cache is not possible")
      #  use_cache=False
    if type(model)!=torch.Tensor:
        model==None
    if model==None:
        if len(grid.shape)==2:
            model=torch.rand(grid.shape)
        else:
            model= torch.rand(grid[0].shape)
   
    
    

    if use_cache:
        NN_grid=torch.from_numpy(np.vstack([grid[i].reshape(-1) for i in range(grid.shape[0])]).T).float()
        if cache_model==None:
            cache_model = torch.nn.Sequential(
                torch.nn.Linear(grid.shape[0], 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 1)
            )
        
        cache_model=matrix_cache_lookup(NN_grid, cache_model, operator, bconds,cache_dir=cache_dir,lambda_bound=100)
        r=create_random_fn(model_randomize_parameter)
        
        cache_model.apply(r)
        
        if len(grid.shape)==2:
            model=cache_model(NN_grid).reshape(grid.shape).detach()
        else:
            model=cache_model(NN_grid).reshape(grid[0].shape).detach()
        
    

    unified_operator = operator_prepare_matrix(operator)

    

    b_prepared = bnd_prepare_matrix(bconds, grid)
    
    if optimizer=='Adam':
        optimizer = torch.optim.Adam([model.requires_grad_()], lr=learning_rate)
    elif optimizer=='SGD':
        optimizer = torch.optim.SGD([model.requires_grad_()], lr=learning_rate)
    elif optimizer=='LBFGS':
        optimizer = torch.optim.LBFGS([model.requires_grad_()], lr=learning_rate)
    else:
        print('Wrong optimizer chosen, optimization was not performed')
        return model
    
    min_loss = matrix_loss(model, grid, unified_operator, b_prepared, lambda_bound=lambda_bound)
    
    # standard NN stuff
    if verbose:
        print('-1 {}'.format(min_loss))
    
    t = 0
    
    last_loss=np.zeros(loss_oscillation_window)+float(min_loss)
    line=np.polyfit(range(loss_oscillation_window),last_loss,1)
    

    def closure():
        nonlocal cur_loss
        optimizer.zero_grad()
        loss =matrix_loss(model, grid, unified_operator, b_prepared, lambda_bound=lambda_bound)
        loss.backward()
        cur_loss = loss.item()
        return loss
    
    stop_dings=0
    t_imp_start=0
    # to stop train proceduce we fit the line in the loss data
    #if line is flat enough 5 times, we stop the procedure
    cur_loss=min_loss
    while stop_dings<=patience:
        optimizer.step(closure)

        last_loss[t%loss_oscillation_window]=cur_loss
        
        if cur_loss<min_loss:
            min_loss=cur_loss
            t_imp_start=t
        if t%loss_oscillation_window==0:
            line=np.polyfit(range(loss_oscillation_window),last_loss,1)
            if abs(line[0]/cur_loss) < eps and t>0:
                stop_dings+=1
                if verbose:
                    print('Oscillation near the same loss')
                    print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
                    solution_print_mat(grid,model,title='Iteration = ' + str(t))
        
        if (t-t_imp_start==no_improvement_patience) and verbose:
            print('No improvement in '+str(no_improvement_patience)+' steps, min_loss = '+str(min_loss))
            t_imp_start=t
            stop_dings+=1
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            solution_print_mat(grid,model,title='Iteration = ' + str(t))
            
        if print_every!=None and (t % print_every == 0) and verbose:
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            solution_print_mat(grid,model,title='Iteration = ' + str(t))

        t += 1
        if t > tmax:
            break

        
    if use_cache or save_always:
        if cache_model==None:
            cache_model = torch.nn.Sequential(
                torch.nn.Linear(grid.shape[0], 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 100),
                torch.nn.Tanh(),
                torch.nn.Linear(100, 1)
            )
        
        NN_grid=torch.from_numpy(np.vstack([grid[i].reshape(-1) for i in range(grid.shape[0])]).T).float()
        optimizer = torch.optim.Adam(cache_model.parameters(), lr=0.001)
        
        model_res=model.reshape(-1,1)
            
        def closure():
            optimizer.zero_grad()
            loss = torch.mean((cache_model(NN_grid)-model_res)**2)
            loss.backward()
            return loss
        
        loss=np.inf
        t=1
        while loss>1e-5 and t<1e5:
            loss = optimizer.step(closure)
            t+=1
            if False:
                print('Retrain from cache t={}, loss={}'.format(t,loss))


        save_model(cache_model,cache_model.state_dict(),optimizer.state_dict(),cache_dir=cache_dir,name=None)
    
        
    return model



def nn_autograd_optimizer(grid, model, operator, bconds, lambda_bound=10,
                            verbose=False, learning_rate=1e-4, eps=1e-5, tmin=1000, tmax=1e5, h=0.001,
                            use_cache=True,cache_dir='../cache/',cache_verbose=False,
                            batch_size=None,save_always=False,lp_par=None,print_every=100,
                            patience=5,loss_oscillation_window=100,no_improvement_patience=1000,
                            model_randomize_parameter=0,optimizer='Adam',abs_loss=None):
    # prepare input data to uniform format 
    
    grid=torch.clone(grid)
    
    full_prepared_operator = operator_prepare_autograd(operator,grid)
    
    prepared_bconds=bnd_prepare_autograd(bconds, grid)
    
    r=create_random_fn(model_randomize_parameter)   
     # use cache if needed
    if use_cache:
        cache_checkpoint,min_loss=cache_lookup_autograd(grid, full_prepared_operator, prepared_bconds,cache_dir=cache_dir
                                                ,nmodels=None,verbose=cache_verbose,lambda_bound=lambda_bound)
        model, optimizer_state= cache_retrain(model,cache_checkpoint,grid,verbose=cache_verbose)
  
        model.apply(r)

    if optimizer=='Adam':
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    elif optimizer=='SGD':
        optimizer = torch.optim.SGD(model.parameters(), lr=learning_rate)
    elif optimizer=='LBFGS':
        optimizer = torch.optim.LBFGS(model.parameters(), lr=learning_rate)
    else:
        print('Wrong optimizer chosen, optimization was not performed')
        return model
    

    if not use_cache:
        min_loss = autograd_loss(model, grid, full_prepared_operator, prepared_bconds, lambda_bound=lambda_bound)
    
    save_cache=False
    
    if min_loss>0.1 or save_always:
        save_cache=True
    
    
    # standard NN stuff
    if verbose:
        print('-1 {}'.format(min_loss))
    
    t = 0
    
    last_loss=np.zeros(loss_oscillation_window)+float(min_loss)
    line=np.polyfit(range(loss_oscillation_window),last_loss,1)
    


    def closure():
        nonlocal cur_loss
        optimizer.zero_grad()
        loss =autograd_loss(model, grid, full_prepared_operator, prepared_bconds, lambda_bound=lambda_bound)
        loss.backward()
        cur_loss = loss.item()
        return loss
    
    stop_dings=0
    t_imp_start=0
    # to stop train proceduce we fit the line in the loss data
    #if line is flat enough 5 times, we stop the procedure
    cur_loss=min_loss
    while stop_dings<=patience:
        optimizer.step(closure)

        last_loss[t%loss_oscillation_window]=cur_loss
        
        if cur_loss<min_loss:
            min_loss=cur_loss
            t_imp_start=t
        if t%loss_oscillation_window==0:
            line=np.polyfit(range(loss_oscillation_window),last_loss,1)
            if abs(line[0]/cur_loss) < eps and t>0:
                stop_dings+=1
                model.apply(r)
                if verbose:
                    print('Oscillation near the same loss')
                    print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
                    solution_print(grid,model,title='Iteration = ' + str(t))
        
        if (t-t_imp_start==no_improvement_patience) and verbose:
            print('No improvement in '+str(no_improvement_patience)+' steps')
            t_imp_start=t
            stop_dings+=1
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            solution_print(grid,model,title='Iteration = ' + str(t))
        
        if abs_loss!=None:
            if cur_loss<abs_loss and verbose:
                print('Absolute value of loss is lower than threshold')
                stop_dings+=1
                print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
                solution_print(grid,model,title='Iteration = ' + str(t))  
     
         
            
        if print_every!=None and (t % print_every == 0) and verbose:
            print(t, cur_loss, line,line[0]/cur_loss, stop_dings)
            solution_print(grid,model,title='Iteration = ' + str(t))

        t += 1
        if t > tmax:
            break
    if (save_cache and use_cache) or save_always:
        save_model(model,model.state_dict(),optimizer.state_dict(),cache_dir=cache_dir,name=None)
    return model



def grid_format_prepare(coord_list, mode='NN'):
    if type(coord_list)==torch.Tensor:
        print('Grid is a tensor, assuming old format, no action performed')
        return coord_list
    if mode=='NN':
        if len(coord_list)==1:
            coord_list=torch.tensor(coord_list)
            grid=coord_list.reshape(-1,1).float()
        else:
            coord_list_tensor=[]
            for item in coord_list:
                if isinstance(item,(np.ndarray)):
                    coord_list_tensor.append(torch.from_numpy(item))
                else:
                    coord_list_tensor.append(item)
            grid=torch.cartesian_prod(*coord_list_tensor).float()
    elif mode=='mat':
        grid = np.meshgrid(*coord_list)
        grid = torch.tensor(grid)
    return grid


def coefficient_unification(grid,operator,mode='NN'):
        for term in operator:
            coeff = term[0]
            if callable(coeff):
                coeff = coeff(grid)
                #coeff = coeff.reshape(-1, 1)
            elif type(coeff) == torch.Tensor:
                if mode == 'NN':
                    if coeff.shape[0]!=grid.shape[0]:
                        coeff = coeff.reshape(-1, 1)
                    if coeff.shape[0]!=grid.shape[0]:
                        print('ERROR: Coefficient shape {} is inconsistent with grid shape {}, it may be because coefficients were pre-computed for matrix grid'.format(coeff.shape,grid.shape))
                elif mode=='mat':
                    if coeff.shape!=grid.shape[1:]:
                        try:
                            coeff = coeff.reshape(grid.shape[1:]).T
                        except Exception:
                            print('ERROR: Something went wrong with the coefficient, it may be because coefficients were pre-computed for NN grid')
            term[0]=coeff
        return operator




def optimization_solver(coord_list, model, operator,bconds,config,mode='NN'):
    grid=grid_format_prepare(coord_list, mode=mode)
    if type(operator)==dict:
        operator=op_dict_to_list(operator)
    operator=coefficient_unification(grid,operator,mode=mode)
    if mode=='NN':
        model=point_sort_shift_solver(grid, model, operator, bconds,
            learning_rate=config.params['Optimizer']['learning_rate'],
            lambda_bound=config.params['Optimizer']['lambda_bound'],
            optimizer=config.params['Optimizer']['optimizer'],
            grid_point_subset=config.params['NN']['grid_point_subset'],
            h=config.params['NN']['h'],
            use_cache=config.params['Cache']['use_cache'],
            cache_dir=config.params['Cache']['cache_dir'],
            cache_verbose=config.params['Cache']['cache_verbose'],
            model_randomize_parameter=config.params['Cache']['model_randomize_parameter'],
            save_always=config.params['Cache']['save_always'],
            batch_size=config.params['NN']['batch_size'],
            lp_par=config.params['NN']['lp_par'],
            verbose=config.params['Verbose']['verbose'],
            print_every=config.params['Verbose']['print_every'],
            eps=config.params['StopCriterion']['eps'], 
            tmin=config.params['StopCriterion']['tmin'], 
            tmax=config.params['StopCriterion']['tmax'], 
            patience=config.params['StopCriterion']['patience'],
            loss_oscillation_window=config.params['StopCriterion']['loss_oscillation_window'],
            no_improvement_patience=config.params['StopCriterion']['no_improvement_patience'])
    if mode=='mat':
        model=matrix_optimizer(grid, model, operator, bconds,
            learning_rate=config.params['Optimizer']['learning_rate'],
            lambda_bound=config.params['Optimizer']['lambda_bound'],
            optimizer=config.params['Optimizer']['optimizer'],
            use_cache=config.params['Cache']['use_cache'],
            cache_dir=config.params['Cache']['cache_dir'],
            cache_verbose=config.params['Cache']['cache_verbose'],
            model_randomize_parameter=config.params['Cache']['model_randomize_parameter'],
            save_always=config.params['Cache']['save_always'],
            lp_par=config.params['Matrix']['lp_par'],
            cache_model=config.params['Matrix']['cache_model'],
            verbose=config.params['Verbose']['verbose'],
            print_every=config.params['Verbose']['print_every'],
            eps=config.params['StopCriterion']['eps'], 
            tmin=config.params['StopCriterion']['tmin'], 
            tmax=config.params['StopCriterion']['tmax'], 
            patience=config.params['StopCriterion']['patience'],
            loss_oscillation_window=config.params['StopCriterion']['loss_oscillation_window'],
            no_improvement_patience=config.params['StopCriterion']['no_improvement_patience'])             
    return model



# -*- coding: utf-8 -*-
"""
Created on Fri Jun  4 18:32:53 2021

@author: user
"""
from copy import copy


# the idea is simple - central difference changes [0]->([1]-[-1])/(2h) (in terms of grid nodes position)
def finite_diff_shift(diff, axis, mode):
    diff_p = copy(diff)
    diff_m = copy(diff)
    if mode == 'central':
        diff_p[axis] = diff_p[axis] + 1
        diff_m[axis] = diff_m[axis] - 1
    elif mode == 'f':
        diff_p[axis] = diff_p[axis] + 1
    elif mode == 'b':
        diff_m[axis] = diff_m[axis] - 1
    return [diff_p, diff_m]


"""
we do the [0]->([1]-[-1])/(2h) transitions to the axes we need
as an example d2u/dxdt
u=[0,0]
u-> du/dx:

[0,0]->([1,0]-[-1,0])/(2h)

du/dx->d2u/dxdt:

[1,0]->([1,1]-[1,-1])/(2h*2tau)

[-1,0]->([-1,1]-[-1,-1])/(2h*2tau)

But we do not want to take signs into account (too complex), so

u-> du/dx:

[0,0]->[[1,0],[-1,0]]

du/dx->d2u/dxdt:

[[1,0],[-1,0]]->[[1,1],[1,-1],[-1,1],[-1,-1]]

Since order is preserved we can compute signs afterwards

"""


def scheme_build(axes, varn, axes_mode):
    order = len(axes)
    finite_diff = []
    direction_list = []
    # we generate [0,0,0,...] for number of variables (varn)
    for i in range(varn):
        finite_diff += [0]
    # just to make this [[0,0,...]]
    finite_diff = [finite_diff]
    # when we increase differential order
    for i in range(order):
        diff_list = []
        for diff in finite_diff:
            # we use [0,0]->[[1,0],[-1,0]] rule for the axis
            if axes_mode == 'central':
                f_diff = finite_diff_shift(diff, axes[i], 'central')
            else:
                f_diff = finite_diff_shift(diff, axes[i], axes_mode[axes[i]])
            if len(diff_list) == 0:
                # and put it to the pool of differentials if it is empty
                diff_list = f_diff
            else:
                # or add to the existing pool
                for diffs in f_diff:
                    diff_list.append(diffs)
        # the we go to the next differential if needed
        finite_diff = diff_list
        direction_list.append(axes_mode[axes[i]])
    return finite_diff, direction_list


# axiluary function to flatten list
flatten_list = lambda t: [item for sublist in t for item in sublist]

"""
From transormations above, we always start from +1 (1)

Every +1 changes to ->[+1,-1] when order of differential rises

[0,0] (+1) ->([1,0]-[-1,0]) ([+1,-1])

Every -1 changes to [-1,+1]

[[1,0],[-1,0]] ([+1,-1])->[[1,1],[1,-1],[-1,1],[-1,-1]] ([+1,-1,-1,+1])

"""


def sign_order(order, mode, h=1 / 2):
    sign_list = [1]
    for i in range(order):
        start_list = []
        for sign in sign_list:
            if mode == 'central':
                start_list.append([sign * (1 / (2 * h)), -sign * (1 / (2 * h))])
            else:
                start_list.append([sign / h, -sign / h])
        sign_list = flatten_list(start_list)
    return sign_list


"""
The following material is presented in illustrative matters
"""

"""
For boundaries we have to use at the one hand - 2nd order scheme, at the other - forward/backward scheme.

Tests show, that we cannot use center scheme even for infinite small h for boundary points, since it gives
function from other than Sobolev space for a given equation.

Thus, everything becomes more serious. Brace yourself, if you go further to the code
"""

"""
2nd order forward difference changes [0]->(a*[2]+b*[1]+c*[0]) (in terms of grid nodes position)
with coefficients a=-3/(2h), b=2/h, c=-1/2h
"""

"""
Lets play with shifts and coeffs separately.
One shift becomes three different instead of 2.
"""
# def second_order_f_shift(diff,axis):
#     diff_1=copy(diff)
#     diff_2=copy(diff)
#     diff_3=copy(diff)
#     diff_3[axis]=diff_3[axis]+2
#     diff_2[axis]=diff_2[axis]+1
#     return [diff_3,diff_2,diff_1]
"""
We also do not want take coeffs into account (too complex), so

u-> du/dx:

[0,0]->[[2,0],[1,0],[0,0]]

du/dx->d2u/dxdt:

[[2,0],[1,0],[0,0]]->[[2,2],[2,1],[2,0],[1,2],[1,1],[1,0],[0,2],[0,1],[0,0]] 

(damn this long)

Since order is preserved we can compute signs afterwards

"""
# def second_order_f_scheme_build(axes,varn):
#     order=len(axes)
#     finite_diff=[]
#     # we generate [0,0,0,...] for number of variables (varn)
#     for i in range(varn):
#         finite_diff+=[0]
#     # just to make this [[0,0,...]]
#     finite_diff=[finite_diff]
#     # when we increase differential order
#     for i in range(order):
#         diff_list=[]
#         for diff in finite_diff:
#             #we use [0,0]->[[1,0],[-1,0]] rule for the axis
#             f_diff=second_order_f_shift(diff,axes[i])
#             if len(diff_list)==0:
#                 #and put it to the pool of differentials if it is empty
#                 diff_list=f_diff
#             else:
#                 # or add to the existing pool
#                 for diffs in f_diff:
#                     diff_list.append(diffs)
#         # the we go to the next differential if needed
#         finite_diff=diff_list
#     return finite_diff


"""
From transormations above, we always start from +1 (1)

Every +1 changes to ->[1,-4,3] when order of differential rises

[0,0] (+1) ->[0,0]->[[2,0],[1,0],[0,0]] ([-3,4,-1])

Every -1 changes to [3a,-4a,a]

[[2,0],[1,0],[0,0]] ([3,-4,1])->[[2,2],[2,1],[2,0],[1,2],[1,1],[1,0],[0,2],[0,1],[0,0]]  ([3,-4,1,-12,16,-4,9,-12,3])

further this should be multiplied by (-1/2h)**order

"""

# def second_order_f_sign_order(order):
#     sign_list=[1]
#     for i in range(order):
#         start_list=[]
#         for sign in sign_list:
#             start_list.append([-3*sign,4*sign,-sign])
#         sign_list=flatten_list(start_list)
#     return sign_list


"""
backward is the same as forward, but with signs inversed
"""

# def second_order_b_shift(diff,axis):
#     diff_1=copy(diff)
#     diff_2=copy(diff)
#     diff_3=copy(diff)
#     diff_3[axis]=diff_3[axis]-2
#     diff_2[axis]=diff_2[axis]-1
#     return [diff_3,diff_2,diff_1]


# def second_order_b_scheme_build(axes,varn):
#     order=len(axes)
#     finite_diff=[]
#     # we generate [0,0,0,...] for number of variables (varn)
#     for i in range(varn):
#         finite_diff+=[0]
#     # just to make this [[0,0,...]]
#     finite_diff=[finite_diff]
#     # when we increase differential order
#     for i in range(order):
#         diff_list=[]
#         for diff in finite_diff:
#             #we use [0,0]->[[1,0],[-1,0]] rule for the axis
#             f_diff=second_order_b_shift(diff,axes[i])
#             if len(diff_list)==0:
#                 #and put it to the pool of differentials if it is empty
#                 diff_list=f_diff
#             else:
#                 # or add to the existing pool
#                 for diffs in f_diff:
#                     diff_list.append(diffs)
#         # the we go to the next differential if needed
#         finite_diff=diff_list
#     return finite_diff

# def second_order_b_sign_order(order):
#     sign_list=[1]
#     for i in range(order):
#         start_list=[]
#         for sign in sign_list:
#             start_list.append([3*sign,-4*sign,sign])
#         sign_list=flatten_list(start_list)
#     return sign_list

"""
The following functions are forward and backward schemes combined
"""


def second_order_shift(diff, axis, mode):
    diff_1 = copy(diff)
    diff_2 = copy(diff)
    diff_3 = copy(diff)
    if mode == 'f':
        diff_3[axis] = diff_3[axis] + 2
        diff_2[axis] = diff_2[axis] + 1
    elif mode == 'b':
        diff_3[axis] = diff_3[axis] - 2
        diff_2[axis] = diff_2[axis] - 1
    else:
        print('Wrong mode')
    return [diff_3, diff_2, diff_1]


def second_order_scheme_build(axes, varn, axes_mode):
    order = len(axes)
    finite_diff = []
    direction_list = []
    # we generate [0,0,0,...] for number of variables (varn)
    for i in range(varn):
        finite_diff += [0]
    # just to make this [[0,0,...]]
    finite_diff = [finite_diff]
    # when we increase differential order
    for i in range(order):
        diff_list = []
        for diff in finite_diff:
            # we use [0,0]->[[1,0],[-1,0]] rule for the axis
            if axes_mode == 'central':
                f_diff = second_order_shift(diff, axes[i], 'central')
            else:
                f_diff = second_order_shift(diff, axes[i], axes_mode[axes[i]])
            if len(diff_list) == 0:
                # and put it to the pool of differentials if it is empty
                diff_list = f_diff
            else:
                # or add to the existing pool
                for diffs in f_diff:
                    diff_list.append(diffs)
        # the we go to the next differential if needed
        finite_diff = diff_list
        direction_list.append(axes_mode[axes[i]])
    return finite_diff, direction_list




def second_order_sign_order(order, mode, h=1 / 2):
    sign_list = [1]
    for i in range(order):
        start_list = []
        for sign in sign_list:
            if mode[i] == 'f':
                start_list.append([3 * (1 / (2 * h)) * sign, -4 * (1 / (2 * h)) * sign, (1 / (2 * h)) * sign])
            elif mode[i] == 'b':
                start_list.append([-3 * (1 / (2 * h)) * sign, 4 * (1 / (2 * h)) * sign, -(1 / (2 * h)) * sign])
        sign_list = flatten_list(start_list)
    return sign_list



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 13 14:13:37 2022

@author: Sashka
"""

from email.policy import default
import json

def read_config(name):
    with open(name, 'r') as config_file:
        config_data = json.load(config_file)
    return config_data


DEFAULT_CONFIG="""
{
"Optimizer": {
"learning_rate":1e-4,
"lambda_bound":10,
"optimizer":"Adam"
},
"Cache":{
"use_cache":true,
"cache_dir":"../cache/",
"cache_verbose":false,
"save_always":false,
"model_randomize_parameter":0
},
"NN":{
"batch_size":null,
"lp_par":null,
"grid_point_subset":["central"],
"h":0.001
},
"Verbose":{
	"verbose":true,
	"print_every":null
},
"StopCriterion":{
"eps":1e-5,
"tmin":1000,
"tmax":1e5 ,
"patience":5,
"loss_oscillation_window":100,
"no_improvement_patience":1000   	
},
"Matrix":{
"lp_par":null,
"cache_model":null
}
}
"""

default_config = json.loads(DEFAULT_CONFIG)

def check_module_name(module_name:str) -> bool:
    """
    check correctness of the 'first' level of config parameter name
    we call it module

    Parameters
    ----------
    module_name: str
        
        first level of a parameter of a custom config

    Returns
    -------
    module_correctness : bool

        true if module presents in 'default' config

    """
    if module_name in default_config.keys():
        return True
    else:
        return False
    


def check_param_name(module_name:str, param_name:str) -> bool:
    """
    check correctness of the 'first' level of config parameter name
    we call it module

    Parameters
    ----------
    module_name: str
        
        first level of a parameter of a custom config

    Returns
    -------
    module_correctness : bool

        true if module presents in 'default' config

    """
    if param_name in default_config[module_name].keys():
        return True
    else:
        return False

'''
We can use old json load. However, it is good to check if the parameters named correctly
So, we make the full 'default' version of the config and load 'non-default' parameters
either from file or using method 'change_parameter'
'''
class Config:
    def __init__(self, *args):
        '''
        We init config with default one

        If there is passed path to a custon config, we try to load it and change
        default parameters


        Parameters
        ----------
        config_path: str, optional
        
            path to a custom config

        Returns
        -------
        self: Config 

            config used in solver.optimization_solver function
        '''



        self.params=default_config
        if len(args)==1:
            try:
                custom_config=read_config(args[0])
            except Exception:
                print('Error reading config. Default config assumed.')
                custom_config=default_config
            for module_name in custom_config.keys():
                if check_module_name(module_name):
                    for param in custom_config[module_name].keys():
                        if check_param_name(module_name,param):
                            self.params[module_name][param]=custom_config[module_name][param]
                        else:
                            print('Wrong parameter name: ok.wrong for {}.{}. Defalut parameters assumed.'.format(module_name,param))
                else: 
                        print('Wrong module name: wrong.maybeok for {}.smth. Defalut parameters assumed.'.format(module_name))

        elif len(args)>1:
            print('Too much initialization args, using default config')
  
       
        
    def set_parameter(self,parameter_string:str,value):
        '''
        We may want to just change defalut condfig parameters manually, without loading
        the .json

        We run checks to see we set them corretly


        Parameters
        ----------
        parameter_string: str
        
        string in format 'module.parameter'

        value: bool,float, int, None

        value for the parameter
        '''
 
        module_name,param=parameter_string.split('.')
        if check_module_name(module_name):
            if check_param_name(module_name,param):
                self.params[module_name][param]=value
            else:
                print('Wrong parameter name: ok.wrong for {}.{}. Defalut parameters assumed.'.format(module_name,param))
        else: 
            print('Wrong module name: wrong.maybeok for {}.smth. Defalut parameters assumed.'.format(module_name))






# -*- coding: utf-8 -*-
"""
Created on Mon Jun 28 22:25:22 2021

@author: Sashka
"""
import numpy as np
import torch
from scipy.spatial import Delaunay


def shift_points(grid, axis, shift):
    """
    Shifts all values of an array 'grid' on a value 'shift' in a direcion of
    axis 'axis', somewhat is equivalent to a np.roll
    """
    grid_shift = grid.clone()
    grid_shift[:, axis] = grid[:, axis] + shift
    return grid_shift


def in_hull(p, hull):
    """
    Test if points in `p` are in `hull`

    `p` should be a `NxK` coordinates of `N` points in `K` dimensions
    `hull` is either a scipy.spatial.Delaunay object or the `MxK` array of the 
    coordinates of `M` points in `K`dimensions for which Delaunay triangulation
    will be computed
    """
    if p.shape[1] > 1:
        if not isinstance(hull, Delaunay):
            hull = Delaunay(hull)

        return hull.find_simplex(p) >= 0
    elif p.shape[1] == 1:
        """
        this one is not a snippet from a stackexchange it does the same 
        but for a 1-D case, which is not covered in a code above
        """
        upbound = torch.max(hull)
        lowbound = torch.min(hull)
        return np.array(((p <= upbound) & (p >= lowbound)).reshape(-1))


def point_typization(grid):
    """
    

    Parameters
    ----------
    grid : torch.Tensor (torch.float64)
        array of a n-D points

    Returns
    -------
    point_type : dict
        dictionary point:type with a points in a 'grid' above
        type may be 'central' - inner point
        and string of 'f' and 'b', where the length of the string 
        is a dimension n
        'f' means that if we add small number to a position of corresponding
        coordinate we stay in the 'hull'
        'b' means that if we subtract small number from o a position 
        of corresponding coordinate we stay in the 'hull'
    """
    direction_list = []
    for axis in range(grid.shape[1]):
        for direction in range(2):
            direction_list.append(in_hull(shift_points(grid, axis, (-1) ** direction * 0.0001), grid))

    direction_list = np.array(direction_list)
    direction_list = np.transpose(direction_list)

    point_type = {}

    for i, point in enumerate(grid):
        if np.all(direction_list[i]):
            point_type[point] = 'central'
        else:
            p_type = ''
            j = 0
            while j < len(direction_list[i]):
                if (j % 2 == 0 and direction_list[i, j]) or (
                        j % 2 == 0 and direction_list[i, j] and direction_list[i, j + 1]):
                    p_type += 'f'
                else:
                    p_type += 'b'
                j += 2
            if grid.shape[-1]==1:
                point_type[point] = 'central'
            else:
                point_type[point] = p_type
    return point_type


def grid_sort(point_type):
    """
    

    Parameters
    ----------
    point_type : dict
        dictionary point:type with a points in a 'grid' above
        type may be 'central' - inner point
        and string of 'f' and 'b', where the length of the string 
        is a dimension n
        'f' means that if we add small number to a position of corresponding
        coordinate we stay in the 'hull'
        'b' means that if we subtract small number from o a position 
        of corresponding coordinate we stay in the 'hull'

    Returns
    -------
    grid_dict : dict
        dictionart type:points list
        basically reversed dictionaty
    """
    point_types = set(point_type.values())
    grid_dict = {}
    for p_type in point_types:
        grid_dict[p_type] = []
    for point in list(point_type.keys()):
        p_type = point_type[point]
        grid_dict[p_type].append(point)
    for p_type in point_types:
        grid_dict[p_type] = torch.stack(grid_dict[p_type])
    return grid_dict



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 24 11:50:12 2021

@author: user
"""
import pickle
import datetime
import torch
import os 
import glob
import numpy as np
from TEDEouS.metrics import point_sort_shift_loss,autograd_loss


def save_model(model,state,optimizer_state,cache_dir='../cache/',name=None):
    if name==None:
        name=str(datetime.datetime.now().timestamp())
    if os.path.isdir(cache_dir):
        torch.save({'model':model, 'model_state_dict': state,
                'optimizer_state_dict': optimizer_state}, cache_dir+name+'.tar')
    else:
        os.mkdir(cache_dir)
        torch.save({'model':model, 'model_state_dict': state,
                'optimizer_state_dict': optimizer_state}, cache_dir+name+'.tar')
    return

def cache_lookup(prepared_grid, operator, bconds, lambda_bound=0.001,cache_dir='../cache/',nmodels=None,verbose=False,norm=None): 
    # looking for all *.tar files in directory cache_dir
    files=glob.glob(cache_dir+'*.tar')
    # if files not found
    if len(files)==0:
        best_checkpoint=None
        min_loss=np.inf
        return best_checkpoint, min_loss
    # at some point we may want to reduce the number of models that are
    # checked for the best in the cache
    if nmodels==None:
        # here we take all files that are in cache
        cache_n=np.arange(len(files))
    else:
        # here we take random nmodels from the cache
        cache_n=np.random.choice(len(files), nmodels, replace=False)
    cache_same_architecture=[]
    min_loss=np.inf
    best_model=0
    best_checkpoint={}
    var = []
    n_vars = []
    for op in operator:
        for term in op:
            var.append(term[4])
    for elt in var:
        nrm = np.sqrt((np.array([-1]) - elt) ** 2)
        for elem in nrm:
            n_vars.append(elem)
    n_vars = int(max(n_vars))
    for i in cache_n:
        file=files[i]
        checkpoint = torch.load(file)
        model=checkpoint['model']
        model.load_state_dict(checkpoint['model_state_dict'])
        # this one for the input shape fix if needed
        # it is taken from the grid shape
        if model[0].in_features != prepared_grid.shape[-1]:
            continue
        try:
            if model[-1].out_features != n_vars:
                continue
        except Exception:
            continue
        # model[0] = torch.nn.Linear(prepared_grid.shape[-1], model[0].out_features)
        # model.eval()
        l=point_sort_shift_loss(model, prepared_grid, operator, bconds, lambda_bound=lambda_bound,norm=norm)      
        if l<min_loss:
            min_loss=l
            best_checkpoint['model']=model
            best_checkpoint['model_state_dict']=model.state_dict()
            best_checkpoint['optimizer_state_dict']=checkpoint['optimizer_state_dict']
            if verbose:
                print('best_model_num={} , loss={}'.format(i,l))
    if best_checkpoint=={}:
        best_checkpoint=None
        min_loss=np.inf
    return best_checkpoint,min_loss
        

def cache_lookup_autograd(grid, operator, bconds, lambda_bound=0.001,cache_dir='../cache/',nmodels=None,verbose=False): 
    # looking for all *.tar files in directory cache_dir
    files=glob.glob(cache_dir+'*.tar')
    # if files not found
    if len(files)==0:
        best_checkpoint=None
        min_loss=np.inf
        return best_checkpoint, min_loss
    # at some point we may want to reduce the number of models that are
    # checked for the best in the cache
    if nmodels==None:
        # here we take all files that are in cache
        cache_n=np.arange(len(files))
    else:
        # here we take random nmodels from the cache
        cache_n=np.random.choice(len(files), nmodels, replace=False)
    cache_same_architecture=[]
    min_loss=np.inf
    best_model=0
    best_checkpoint={}
    for i in cache_n:
        file=files[i]
        checkpoint = torch.load(file)
        model=checkpoint['model']
        model.load_state_dict(checkpoint['model_state_dict'])
        # this one for the input shape fix if needed
        # it is taken from the grid shape
        if model[0].in_features!=grid.shape[-1]:
            continue
        #     model[0]=torch.nn.Linear(prepared_grid.shape[-1],model[0].out_features)
        # model.eval()
        l=autograd_loss(model, grid, operator, bconds, lambda_bound=lambda_bound)     
        if l<min_loss:
            min_loss=l
            best_checkpoint['model']=model
            best_checkpoint['model_state_dict']=model.state_dict()
            best_checkpoint['optimizer_state_dict']=checkpoint['optimizer_state_dict']
            if verbose:
                print('best_model_num={} , loss={}'.format(i,l))
    if best_checkpoint=={}:
        best_checkpoint=None
        min_loss=np.inf
    return best_checkpoint,min_loss
        


def cache_retrain(model,cache_checkpoint,grid,verbose=False):
    # do nothing if cache is empty
    if cache_checkpoint==None:
        optimizer_state=None
        return model,optimizer_state
    # if models have the same structure use the cache model state
    if str(cache_checkpoint['model']) == str(model):
        model=cache_checkpoint['model']
        model.load_state_dict(cache_checkpoint['model_state_dict'])
        model.eval()
        optimizer_state=cache_checkpoint['optimizer_state_dict']
        if verbose:
            print('Using model from cache')
    # else retrain the input model using the cache model 
    else:
        optimizer_state=None
        model_state=None
        optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
        cache_model=cache_checkpoint['model']
        cache_model.load_state_dict(cache_checkpoint['model_state_dict'])
        cache_model.eval()
        loss = torch.mean(torch.square(cache_model(grid)-model(grid)))
        t=1
        def closure():
            optimizer.zero_grad()
            loss = torch.mean((cache_model(grid)-model(grid))**2)
            loss.backward()
            return loss
        t=1
        while loss>1e-5 and t<1e5:
            optimizer.step(closure)
            loss = torch.mean(torch.square(cache_model(grid)-model(grid)))
            t+=1
            if verbose:
                print('Retrain from cache t={}, loss={}'.format(t,loss))
    return model, optimizer_state


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 22 21:31:03 2021

@author: user
"""
import torch

from TEDEouS.points_type import point_typization
from TEDEouS.points_type import grid_sort
from TEDEouS.finite_diffs import scheme_build
from TEDEouS.finite_diffs import sign_order
from TEDEouS.finite_diffs import second_order_scheme_build
from TEDEouS.finite_diffs import second_order_sign_order
import numpy as np

"""
Functions here serve to make all possible forms of input into an inner representation of solver
"""


def grid_prepare(grid):
    """
    In the practice we want to apply different finite difference schemes to a
    inner and boundary points. This one is axilluary function that allow to
    reorder grid nodes with respect to the point types - inner points first
    then boundary points of different types

    Parameters
    ----------
    grid : torch.Tensor (torch.float64)


    Returns
    -------
    sorted_grid : torch.Tensor (torch.float64)

    """
    point_type = point_typization(grid)

    grid_dict = grid_sort(point_type)

    sorted_grid = torch.cat(list(grid_dict.values()))
    
    
    return sorted_grid,grid_dict,point_type


def operator_unify(operator):
    """
    I just was annoyed adding additional square brackets to the operators.
    This one allows to make operator form simpler.

    Parameters
    ----------
    operator : list
        Operator in form ... .

    Returns
    -------
    unified_operator : list
        DESCRIPTION.

    """
    unified_operator = []
    for term in operator:
        const = term[0]
        vars_set = term[1]
        power = term[2]
        variables=[]
        if len(term)==4:
            variables = term[3]
        elif isinstance(power,(int,float)):
            variables=0
        elif type(power)==list:
            for _ in range(len(power)):
                variables.append(0)
        if variables is None:
            if type(power) is list:
                unified_operator.append([const, vars_set, power,0])
            else:
                unified_operator.append([const, [vars_set], [power],[0]])
        else:
            if type(power) is list:
                unified_operator.append([const, vars_set, power,variables])
            else:
                unified_operator.append([const, [vars_set], [power],[variables]])
        # if type(power) is list:
        #         unified_operator.append([const, vars_set, power])
        # else:
        #         unified_operator.append([const, [vars_set], [power]])
    return unified_operator


def operator_to_type_op(unified_operator, nvars, axes_scheme_type, h=1 / 2,inner_order=1, boundary_order=2):
    """
    Function serves applying different schemes to a different point types for
    entire operator

    Parameters
    ----------
    unified_operator : list
        operator in a proper form
    nvars : int
        Dimensionality of the problem.
    axes_scheme_type : string
        'central' or combination of 'f' and 'b'.
    h : float, optional
        Derivative precision parameter (to simplify h in the expression
                                        (f(x+h)-f(x))/h). The default is 1/2.
    boundary_order : int, optional
        Order of finite difference scheme taken at the domain boundary points. 
        The default is 1.

    Returns
    -------
    fin_diff_op : list
        list, where the conventional operator changed to steps and signs
        (see scheme_build function description).

    """
    fin_diff_op = []
    for term in unified_operator:
        fin_diff_list = []
        s_order_list = []
        const = term[0]
        vars_set = term[1]
        power = term[2]
        variables = term[3]
        for k, term in enumerate(vars_set):
            # None is for case where we need the fuction without derivatives
            if term != [None]:
                if axes_scheme_type == 'central':
                    if inner_order==1:
                        scheme, direction_list = scheme_build(term, nvars, 'central')
                        s_order = sign_order(len(term), 'central', h=h)
                    elif inner_order==2:
                        scheme, direction_list = second_order_scheme_build(term, nvars, 'central')
                        s_order = second_order_sign_order(len(term), 'central', h=h)
                else:
                    if boundary_order == 1:
                        scheme, direction_list = scheme_build(term, nvars, axes_scheme_type)
                        s_order = sign_order(len(term), direction_list, h=h)
                    elif boundary_order == 2:
                        scheme, direction_list = second_order_scheme_build(term, nvars, axes_scheme_type)
                        s_order = second_order_sign_order(len(term), direction_list, h=h)
            else:
                scheme = [None]
                s_order = [1]
            fin_diff_list.append(scheme)
            s_order_list.append(s_order)
        fin_diff_op.append([const, fin_diff_list, s_order_list, power,variables])
    return fin_diff_op


def shift_points(grid, axis, shift):
    """
    Shifts all values of an array 'grid' on a value 'shift' in a direcion of
    axis 'axis', somewhat is equivalent to a np.roll
    """
    grid_shift = grid.clone()
    grid_shift[:, axis] = grid[:, axis] + shift
    return grid_shift


def finite_diff_scheme_to_grid_list(finite_diff_scheme, grid, h=0.001):
    """
    Axiluary function that converts integer grid steps in term described in
    finite-difference scheme to a grids with shifted points, i.e.
    from field (x,y) -> (x,y+h).

    Parameters
    ----------
    finite_diff_scheme : list
        operator_to_type_op one term
    grid : torch.Tensor
    h : float
        derivative precision parameter. The default is 0.001.

    Returns
    -------
    s_grid_list : list
        list, where the the steps and signs changed to grid and signs
    """
    s_grid_list = []
    for i, shifts in enumerate(finite_diff_scheme):
        s_grid = grid
        for j, axis in enumerate(shifts):
            if axis != 0:
                s_grid = shift_points(s_grid, j, axis * h)
        s_grid_list.append(s_grid)
    return s_grid_list




def type_op_to_grid_shift_op(fin_diff_op, grid, h=0.001, true_grid=None):
    """
    Converts operator to a grid_shift form. Includes term coefficient
    conversion.
    Coeff may be integer, function or array, last two are mapped to a 
    subgrid that corresponds point type

    Parameters
    ----------
    fin_diff_op : list
        operator_to_type_op result.
    grid : torch.Tensor
        grid with sotred nodes (see grid_prepare)
    h : float
        derivative precision parameter. The default is 0.001.
    true_grid : TYPE, optional
        initial grid for coefficient in form of torch.Tensor mapping

    Returns
    -------
    shift_grid_op : list
        final form of differential operator used in the algorithm for single 
        grid type
    """
    shift_grid_op = []
    for term1 in fin_diff_op:
        shift_grid_list = []
        coeff1 = term1[0]
        if type(coeff1) == int or type(coeff1) == float:
            coeff = coeff1
        elif callable(coeff1):
            coeff = coeff1(grid)
            coeff = coeff.reshape(-1, 1)
        elif type(coeff1) == torch.Tensor:
            if true_grid != None:
                pos = bndpos(true_grid, grid)
            else:
                pos = bndpos(grid, grid)
            coeff = coeff1[pos].reshape(-1, 1)
        
        finite_diff_scheme = term1[1]
        s_order = term1[2]
        power = term1[3]
        variables = term1[4]
        for k, term in enumerate(finite_diff_scheme):
            if term != [None]:
                grid_op = finite_diff_scheme_to_grid_list(term, grid, h=h)
            else:
                grid_op = [grid]
            shift_grid_list.append(grid_op)
        shift_grid_op.append([coeff, shift_grid_list, s_order, power,variables])
    return shift_grid_op


def apply_all_operators(operator, grid_dict, h=0.001, subset=None, true_grid=None):
    """
    

    Parameters
    ----------
    operator : list
        operator_unify result.
    grid : torch.Tensor
        grid with sotred nodes (see grid_prepare)
    h : float
        derivative precision parameter. The default is 0.001.
    subset : list, optional
        grid subsets used for the operator ,e.g. ['central','fb','ff']
    true_grid : TYPE, optional
        initial grid for coefficient in form of torch.Tensor mapping

    Returns
    -------
    operator_list :  list
        final form of differential operator used in the algorithm for subset 
        grid types

    """
    operator_list = []
    # nvars = grid.shape[1]
    # point_type = point_typization(grid)
    # grid_dict = grid_sort(point_type)
    nvars =list(grid_dict.values())[0].shape[-1]
    a = operator_unify(operator)
    for operator_type in list(grid_dict.keys()):
        if subset == None or operator_type in subset:
            b = operator_to_type_op(a, nvars, operator_type, h=h)
            c = type_op_to_grid_shift_op(b, grid_dict[operator_type], h=h, true_grid=true_grid)
            operator_list.append(c)
    return operator_list


def operator_prepare(op, grid_dict, subset=['central'], true_grid=None, h=0.001):
    """
    Changes the operator in conventional form to the input one
    
    Parameters
    ----------
    op : list
        operator in conventional form.
    grid : torch.Tensor
        grid with sotred nodes (see grid_prepare)
    h : float
        derivative precision parameter. The default is 0.001.
    subset : list, optional
        grid subsets used for the operator ,e.g. ['central','fb','ff']
    true_grid : torch.Tensor, optional
        initial grid for coefficient in form of torch.Tensor mapping
    Returns
    -------
    operator_list :  list
        final form of differential operator used in the algorithm for subset 
        grid types

    """
    if type(op)==dict:
        op=op_dict_to_list(op)
    op1 = operator_unify(op)
    prepared_operator = apply_all_operators(op1, grid_dict, subset=subset, true_grid=true_grid, h=h)
    return prepared_operator





def op_dict_to_list(opdict):
    return list([list(term.values()) for term in opdict.values()])

def closest_point(grid,target_point):
    min_dist=np.inf
    pos=0
    min_pos=0
    for point in grid:
        dist=torch.linalg.norm(point-target_point)
        if dist<min_dist:
            min_dist=dist
            min_pos=pos
        pos+=1
    return min_pos


def bndpos(grid, bnd):
    """
    
    Returns the position of the boundary points on the grid
    
    Parameters
    ----------
    grid : torch.Tensor
        grid for coefficient in form of torch.Tensor mapping
    bnd : torch.Tensor
        boundary

    Returns
    -------
    bndposlist : list (int)
        positions of boundaty points in grid

    """
    if grid.shape[0]==1:
        grid=grid.reshape(-1,1)
    bndposlist = []
    grid = grid.double()
    if type(bnd) == np.array:
        bnd = torch.from_numpy(bnd).double()
    else:
        bnd = bnd.double()
    for point in bnd:
        try:
            pos = int(torch.where(torch.all(torch.isclose(grid, point), dim=1))[0])
        except Exception:
            pos=closest_point(grid,point)
        bndposlist.append(pos)
    return bndposlist


def bnd_unify(bconds):
    """
    Serves to add None instead of empty operator

    Parameters
    ----------
    bconds : list
        
        boundary in conventional form (see examples)

    Returns
    -------
    unified_bconds : list
        
        boundary in input-friendly form

    """
    if bconds==None:
        return None
    unified_bconds = []
    for bcond in bconds:
        if len(bcond) == 2:
            unified_bconds.append([bcond[0], None, bcond[1]])
        elif len(bcond) == 3:
            unified_bconds.append(bcond)
    return unified_bconds


def bnd_prepare(bconds,grid,grid_dict, h=0.001):
    """
    

    Parameters
    ----------
    bconds : list
        boundary in conventional form (see examples)
    grid : torch.Tensor
        grid with sotred nodes (see grid_prepare)
    h : float
        derivative precision parameter. The default is 0.001.

    Returns
    -------
    prepared_bnd : list
        
        boundary in input form

    """
    bconds = bnd_unify(bconds)
    if bconds==None:
        return None
    prepared_bnd = []
    for bcond in bconds:
        b_coord = bcond[0]
        bop = bcond[1]
        bval = bcond[2]
        bpos = bndpos(grid, b_coord)
        if bop == [[1, [None], 1]]:
            bop = None
        if bop != None:
            if type(bop)==dict:
                bop=op_dict_to_list(bop)
            bop1 = operator_unify(bop)
            bop2 = apply_all_operators(bop1, grid_dict, h=h)
        else:
            bop2 = None
        prepared_bnd.append([bpos, bop2, bval])

    return prepared_bnd


def grid_intersect(t1, t2):
    t1=list(t1.cpu().numpy())
    t2=list(t2.cpu().numpy())    
    t1_set=[]
    t2_set=[]
    for item in t1:
        t1_set.append(tuple(item))
    for item in t2:
        t2_set.append(tuple(item))
    t1_set=set(t1_set)
    t2_set=set(t2_set)
    intersect=t1_set.intersection(t2_set)
    intersect=list(intersect)
    for i in range(len(intersect)):
        intersect[i]=list(intersect[i])
    intersect=torch.Tensor(intersect)
    return intersect

    
def batch_bconds_transform(batch_grid,bconds):
    bconds = bnd_unify(bconds)
    batch_bconds=[]
    for bcond in bconds:
        b_coord = bcond[0]
        bop = bcond[1]
        bval = bcond[2]
        grid_proj=grid_intersect(b_coord, batch_grid)
        if len(grid_proj)>0:
            proj_pos=bndpos(b_coord, grid_proj)
            bval=bval[proj_pos]
            batch_bconds.append([grid_proj,bop,bval])
    if len(batch_bconds)==0:
        batch_bconds=None
    return batch_bconds



def bnd_prepare_matrix(bconds,grid):
    prepared_bconds=[]
    for bnd in bconds:
        if len(bnd)==2:
            bpts=bnd[0]
            bop=None
            bval=bnd[1]
        else:
            bpts=bnd[0]
            bop=bnd[1]
            bval=bnd[2]
        bpos=[]
        # bpos=bndpos(grid,bpts)
        for pt in bpts:
            if grid.shape[0]==1:
                point_pos=(torch.tensor(bndpos(grid,pt)),)
            else:
                prod=(torch.zeros_like(grid[0])+1).bool()
                for axis in range(grid.shape[0]):
                    axis_intersect=torch.isclose(pt[axis].float(),grid[axis].float())
                    prod*=axis_intersect
                point_pos=torch.where(prod==True)
            bpos.append(point_pos)
        if type(bop)==dict:
            bop=op_dict_to_list(bop)
        if bop!=None:
            bop = operator_unify(bop)
        prepared_bconds.append([bpos,bop,bval])
    return prepared_bconds

def operator_prepare_matrix(operator):
    if type(operator) == dict:
        operator = op_dict_to_list(operator)
    unified_operator = operator_unify(operator)
    return unified_operator


def expand_coeffs_autograd(op,grid):
    autograd_op=[]
    for term in op:
        coeff1 = term[0]
        if type(coeff1) == int:
            coeff = coeff1
        elif callable(coeff1):
            coeff = coeff1(grid)
            coeff = coeff.reshape(-1,1)
        elif type(coeff1) == torch.Tensor:
            coeff = coeff1.reshape(-1,1)
        prod = term[1]
        power = term[2]
        autograd_op.append([coeff, prod, power])
    return autograd_op

def operator_prepare_autograd(op,grid):
    """
    Changes the operator in conventional form to the input one
    
    Parameters
    ----------
    op : list
        operator in conventional form.
    Returns
    -------
    operator_list :  list
        final form of differential operator used in the algorithm 

    """
    if type(op)==dict:
        op=op_dict_to_list(op)
    unified_operator = operator_unify(op)
        
    prepared_operator=expand_coeffs_autograd(unified_operator,grid)
    
    return prepared_operator


def bnd_prepare_autograd(bconds,grid):
    """
    

    Parameters
    ----------
    bconds : list
        boundary in conventional form (see examples)
    grid : torch.Tensor
        grid with sotred nodes (see grid_prepare)
    h : float
        derivative precision parameter. The default is 0.001.

    Returns
    -------
    prepared_bnd : list
        
        boundary in input form

    """
    bconds = bnd_unify(bconds)
    if bconds==None:
        return None
    prepared_bnd = []
    for bcond in bconds:
        b_coord = bcond[0]
        bop = bcond[1]
        bval = bcond[2]
        bpos = bndpos(grid, b_coord)
        if bop == [[1, [None], 1]]:
            bop = None
        if bop != None:
            if type(bop)==dict:
                bop=op_dict_to_list(bop)
            bop1 = operator_unify(bop)
        else:
            bop1 = None
        prepared_bnd.append([bpos, bop1, bval])
    return prepared_bnd




import torch
import numpy as np
from TEDEouS.input_preprocessing import grid_prepare, bnd_prepare, operator_prepare,batch_bconds_transform
from TEDEouS.points_type import grid_sort

def take_derivative_shift_op(model, term):
    """
    Axiluary function serves for single differential operator resulting field
    derivation

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    term : TYPE
        differential operator in conventional form.

    Returns
    -------
    der_term : torch.Tensor
        resulting field, computed on a grid.

    """
    # it is may be int, function of grid or torch.Tensor
    coeff = term[0]
    # this one contains shifted grids (see input_preprocessing module)
    shift_grid_list = term[1]
    # signs corresponding to a grid
    s_order_norm_list = term[2]
    # float that represents power of the differential term
    power = term[3]
    # number of variables in equation
    variables = term[4]
    # initially it is an ones field
    der_term = (torch.zeros_like(model(shift_grid_list[0][0])[0:,0]) + 1).reshape(-1,1)
    
    for j, scheme in enumerate(shift_grid_list):
        # every shift in grid we should add with correspoiding sign, so we start
        # from zeros
        grid_sum = torch.zeros_like(model(scheme[0]))[0:,0].reshape(-1,1) #почему схема от 0?
        for k, grid in enumerate(scheme):
            # and add grid sequentially
            grid_sum += (model(grid)[0:,variables[j]]).reshape(-1,1) * s_order_norm_list[j][k]
            # Here we want to apply differential operators for every term in the product
        der_term = der_term * grid_sum ** power[j]
    der_term = coeff * der_term

         
    return der_term


def apply_const_shift_operator(model, operator):
    """
    Deciphers equation in a single grid subset to a field.

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    operator : list
        Single (len(subset)==1) operator in input form. See 
        input_preprocessing.operator_prepare()

    Returns
    -------
    total : torch.Tensor

    """
    for term in operator:
        dif = take_derivative_shift_op(model, term)
        try:
            total += dif
        except NameError:
            total = dif
    return total


def apply_operator_set(model, operator_set):
    """
    Deciphers equation in a whole grid to a field.

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    operator : list
        Multiple (len(subset)>=1) operators in input form. See 
        input_preprocessing.operator_prepare()

    Returns
    -------
    total : torch.Tensor

    """
    field_part = []
    for operator in operator_set:
        field_part.append(apply_const_shift_operator(model, operator))
    field_part = torch.cat(field_part)
    return field_part


flatten_list = lambda t: [item for sublist in t for item in sublist]

def lp_norm(*arg,p=2,normalized=False,weighted=False):
    if weighted==True and len(arg)==1:
        print('No grid is passed, using non-weighted norm')
        weighted=False
    if len(arg)==2:
        grid=arg[0]
        mat=arg[1]
    elif len(arg)==1:
        mat=arg[0]
        grid=None
    else:
        print('Something went wrong, passed more than two arguments')
        return
    grid_prod=1
    if weighted:
        for i in range(grid.shape[-1]):
            grid_prod*=grid[:,i]
    if p>1: 
        if not weighted and not normalized:
             norm=torch.mean((mat) ** p)
        elif not weighted and normalized:
            norm=torch.pow(torch.mean((mat) ** p),1/p)
        elif weighted and not normalized:
            norm=torch.mean(grid_prod*(mat) ** p)
        elif weighted and normalized:
            norm=torch.pow(torch.mean(grid_prod*(mat) ** p),1/p)
    elif p==1:
        if not weighted:
             norm=torch.mean(torch.abs(mat))
        elif weighted:
            norm=torch.mean(grid_prod*torch.abs(mat))  
    return norm


def point_sort_shift_loss(model, grid, operator_set, bconds, lambda_bound=10,norm=None):

    op = apply_operator_set(model, operator_set)
    if bconds==None:
        loss = torch.mean((op) ** 2)
        return loss
    
    true_b_val_list = []
    b_val_list = []
    b_pos_list = []

    # we apply no  boundary conditions operators if they are all None

    simpleform = False
    for bcond in bconds:
        if bcond[1] == None:
            simpleform = True
        if bcond[1] != None:
            simpleform = False
            break
    if simpleform:
        for bcond in bconds:
            b_pos_list.append(bcond[0])

            if len(bcond[2]) == bcond[2].shape[-1]:
                true_boundary_val = bcond[2].reshape(-1,1)
            else: 
                true_boundary_val = bcond[2]

            true_b_val_list.append(true_boundary_val)
        true_b_val = torch.cat(true_b_val_list)
        b_op_val = model(grid)
        b_val = b_op_val[flatten_list(b_pos_list)]
    # or apply differential operator first to compute corresponding field and
    else:
        for bcond in bconds:
            b_pos = bcond[0]
            b_pos_list.append(bcond[0])
            b_cond_operator = bcond[1]
            
            if len(bcond[2]) == bcond[2].shape[-1]:
                true_boundary_val = bcond[2].reshape(-1,1)
            else: 
                true_boundary_val = bcond[2]
            true_b_val_list.append(true_boundary_val)
            if b_cond_operator == None or b_cond_operator == [[1, [None], 1]]:
                b_op_val = model(grid)
            else:
                b_op_val = apply_operator_set(model, b_cond_operator)
            # take boundary values
            b_val_list.append(b_op_val[b_pos])
        true_b_val = torch.cat(true_b_val_list)
        b_val = torch.cat(b_val_list)

    """
    actually, we can use L2 norm for the operator and L1 for boundary
    since L2>L1 and thus, boundary values become not so signifnicant, 
    so the NN converges faster. On the other hand - boundary conditions is the
    crucial thing for all that stuff, so we should increase significance of the
    coundary conditions
    """
    if norm==None:
        op_weigthed=False
        op_normalized=False
        op_p=2
        b_weigthed=False
        b_normalized=False
        b_p=2
    else:
        op_weigthed=norm['operator_weighted']
        op_normalized=norm['operator_normalized']
        op_p=norm['operator_p']
        b_weigthed=norm['boundary_weighted']
        b_normalized=norm['boundary_weighted']
        b_p=norm['boundary_p']
    
    loss = lp_norm(grid[:len(op)],op,weighted=op_weigthed,normalized=op_normalized,p=op_p) + \
    lambda_bound * lp_norm(grid[flatten_list(b_pos_list)],b_val - true_b_val,p=b_p,weighted=b_weigthed,normalized=b_normalized)
    
    return loss

def point_sort_shift_loss_batch(model, prepared_grid, point_type, operator, bconds,subset=['central'], lambda_bound=10,batch_size=32,h=0.001,norm=None):
    permutation = torch.randperm(prepared_grid.size()[0])
    loss=0
    batch_num=0
    for i in range(0,prepared_grid.size()[0], batch_size):
        indices = permutation[i:i+batch_size]
        if len(indices)<5:
            continue
        # batch= grid[indices]
        
        # batch_grid = grid_prepare(batch)
        batch_grid=prepared_grid[indices]


        batch_types=np.array(list(point_type.values()))[indices.tolist()]
        
        batch_type=dict(zip(batch_grid, batch_types))
        
        batch_dict=grid_sort(batch_type)
        batch_bconds=batch_bconds_transform(batch_grid,bconds)
        batch_bconds = bnd_prepare(batch_bconds, batch_grid,batch_dict, h=h)

        batch_operator = operator_prepare(operator, batch_dict, subset=subset, true_grid=prepared_grid[indices], h=h)
        
        
        loss+= point_sort_shift_loss(model, batch_grid, batch_operator, batch_bconds, lambda_bound=lambda_bound,norm=norm)
        batch_num+=1
    loss=1/batch_num*loss
    return loss


def compute_operator_loss(grid, model, operator, bconds, grid_point_subset=['central'], lambda_bound=10,h=0.001):
    prepared_grid = grid_prepare(grid)
    bconds = bnd_prepare(bconds, prepared_grid, h=h)
    operator = operator_prepare(operator, prepared_grid, subset=grid_point_subset, true_grid=grid, h=h)
    loss = point_sort_shift_loss(model, prepared_grid, operator, bconds, lambda_bound=lambda_bound)
    loss=float(loss.float())
    return loss


def derivative_1d(model,grid):
    # print('1d>2d')
    u=model.reshape(-1)
    x=grid.reshape(-1)
    
    # du_forward = (u-torch.roll(u, -1)) / (x-torch.roll(x, -1))
    
    # du_backward = (torch.roll(u, 1) - u) / (torch.roll(x, 1) - x)
    du =  (torch.roll(u, 1) - torch.roll(u, -1))/(torch.roll(x, 1)-torch.roll(x, -1))
    du[0] = (u[0]-u[1])/(x[0]-x[1])
    du[-1] = (u[-1]-u[-2])/(x[-1]-x[-2])
    
    du=du.reshape(model.shape)
    
    return du


def derivative(u_tensor, h_tensor, axis, scheme_order=1, boundary_order=1):
    #print('shape=',u_tensor.shape)
    if (u_tensor.shape[0]==1):
        du=derivative_1d(u_tensor,h_tensor)
        return du

    u_tensor = torch.transpose(u_tensor, 0, axis)
    h_tensor = torch.transpose(h_tensor, 0, axis)
    
    
    if scheme_order==1:
        du_forward = (-torch.roll(u_tensor, -1) + u_tensor) / \
                     (-torch.roll(h_tensor, -1) + h_tensor)
    
        du_backward = (torch.roll(u_tensor, 1) - u_tensor) / \
                      (torch.roll(h_tensor, 1) - h_tensor)
        du = (1 / 2) * (du_forward + du_backward)
    
    # dh=h_tensor[0,1]-h_tensor[0,0]
    
    if scheme_order==2:
        u_shift_down_1 = torch.roll(u_tensor, 1)
        u_shift_down_2 = torch.roll(u_tensor, 2)
        u_shift_up_1 = torch.roll(u_tensor, -1)
        u_shift_up_2 = torch.roll(u_tensor, -2)
        
        h_shift_down_1 = torch.roll(h_tensor, 1)
        h_shift_down_2 = torch.roll(h_tensor, 2)
        h_shift_up_1 = torch.roll(h_tensor, -1)
        h_shift_up_2 = torch.roll(h_tensor, -2)
        
        h1_up=h_shift_up_1-h_tensor
        h2_up=h_shift_up_2-h_shift_up_1
        
        h1_down=h_tensor-h_shift_down_1
        h2_down=h_shift_down_1-h_shift_down_2
       
        a_up=-(2*h1_up+h2_up)/(h1_up*(h1_up+h2_up))
        b_up=(h2_up+h1_up)/(h1_up*h2_up)
        c_up=-h1_up/(h2_up*(h1_up+h2_up))
        
        a_down=(2*h1_down+h2_down)/(h1_down*(h1_down+h2_down))
        b_down=-(h2_down+h1_down)/(h1_down*h2_down)
        c_down=h1_down/(h2_down*(h1_down+h2_down))
        
        du_forward=a_up*u_tensor+b_up*u_shift_up_1+c_up*u_shift_up_2
        du_backward=a_down*u_tensor+b_down*u_shift_down_1+c_down*u_shift_down_2
        du = (1 / 2) * (du_forward + du_backward)
        
        
    if boundary_order==1:
        if scheme_order==1:
            du[:, 0] = du_forward[:, 0]
            du[:, -1] = du_backward[:, -1]
        elif scheme_order==2:
            du_forward = (-torch.roll(u_tensor, -1) + u_tensor) / \
                         (-torch.roll(h_tensor, -1) + h_tensor)
        
            du_backward = (torch.roll(u_tensor, 1) - u_tensor) / \
                          (torch.roll(h_tensor, 1) - h_tensor)
            du[:, 0] = du_forward[:, 0]
            du[:, 1] = du_forward[:, 1]
            du[:, -1] = du_backward[:, -1]
            du[:, -2] = du_backward[:, -2]
    elif boundary_order==2:
        if scheme_order==2:
             du[:, 0] = du_forward[:, 0]
             du[:, 1] = du_forward[:, 1]
             du[:, -1] = du_backward[:, -1]
             du[:, -2] = du_backward[:, -2]
        elif scheme_order==1:
            u_shift_down_1 = torch.roll(u_tensor, 1)
            u_shift_down_2 = torch.roll(u_tensor, 2)
            u_shift_up_1 = torch.roll(u_tensor, -1)
            u_shift_up_2 = torch.roll(u_tensor, -2)
            
            h_shift_down_1 = torch.roll(h_tensor, 1)
            h_shift_down_2 = torch.roll(h_tensor, 2)
            h_shift_up_1 = torch.roll(h_tensor, -1)
            h_shift_up_2 = torch.roll(h_tensor, -2)
            
            h1_up=h_shift_up_1-h_tensor
            h2_up=h_shift_up_2-h_shift_up_1
            
            h1_down=h_tensor-h_shift_down_1
            h2_down=h_shift_down_1-h_shift_down_2
           
       
            a_up=-(2*h1_up+h2_up)/(h1_up*(h1_up+h2_up))
            b_up=(h2_up+h1_up)/(h1_up*h2_up)
            c_up=-h1_up/(h2_up*(h1_up+h2_up))
            
            a_down=(2*h1_up+h2_up)/(h1_down*(h1_down+h2_down))
            b_down=-(h2_down+h1_down)/(h1_down*h2_down)
            c_down=h1_down/(h2_down*(h1_down+h2_down))
        
            
            du_forward=a_up*u_tensor+b_up*u_shift_up_1+c_up*u_shift_up_2
            du_backward=a_down*u_tensor+b_down*u_shift_down_1+c_down*u_shift_down_2
            du[:, 0] = du_forward[:, 0]
            du[:, -1] = du_backward[:, -1]
            
    du = torch.transpose(du, 0, axis)

    return du


def take_matrix_diff_term(model, grid, term):
    """
    Axiluary function serves for single differential operator resulting field
    derivation

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    term : TYPE
        differential operator in conventional form.

    Returns
    -------
    der_term : torch.Tensor
        resulting field, computed on a grid.

    """
    # it is may be int, function of grid or torch.Tensor
    coeff = term[0]
    # this one contains product of differential operator
    operator_product = term[1]
    # float that represents power of the differential term
    power = term[2]
    # initially it is an ones field
    der_term = torch.zeros_like(model) + 1
    for j, scheme in enumerate(operator_product):
        prod=model
        if scheme!=[None]:
            for axis in scheme:
                if axis is None:
                    continue
                h = grid[axis]
                prod=derivative(prod, h, axis, scheme_order=1, boundary_order=1)
        der_term = der_term * prod ** power[j]
    if callable(coeff) is True:
        der_term = coeff(grid) * der_term
    else:
        der_term = coeff * der_term
    return der_term

def apply_matrix_diff_operator(model,grid, operator):
    """
    Deciphers equation in a single grid subset to a field.

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    operator : list
        Single (len(subset)==1) operator in input form. See
        input_preprocessing.operator_prepare()

    Returns
    -------
    total : torch.Tensor

    """
    for term in operator:
        dif = take_matrix_diff_term(model, grid, term)
        try:
            total += dif
        except NameError:
            total = dif
    return total



def apply_matrix_bcond_operator(model,grid,b_prepared):
    true_b_val_list = []
    b_val_list = []

    # we apply no  boundary conditions operators if they are all None

    simpleform = False
    for bcond in b_prepared:
        if bcond[1] == None:
            simpleform = True
        if bcond[1] != None:
            simpleform = False
            break
    if simpleform:
        for bcond in b_prepared:
            b_pos = bcond[0]
            true_boundary_val = bcond[2]
            for position in b_pos:
                if grid.dim() == 1 or min(grid.shape) == 1:
                    b_val_list.append(model[:,position])
                else:
                    b_val_list.append(model[position])
            true_b_val_list.append(true_boundary_val)
        b_val=torch.cat(b_val_list)
        true_b_val=torch.cat(true_b_val_list)
    # or apply differential operatorn first to compute corresponding field and
    else:
        for bcond in b_prepared:
            b_pos = bcond[0]
            b_cond_operator = bcond[1]
            true_boundary_val = bcond[2]
            if b_cond_operator == None or b_cond_operator == [[1, [None], 1]]:
                b_op_val = model
            else:
                b_op_val = apply_matrix_diff_operator(model,grid,b_cond_operator)
            # take boundary values
            for position in b_pos:
                if grid.dim() == 1 or min(grid.shape) == 1:
                    b_val_list.append(b_op_val[:, position])
                else:
                    b_val_list.append(b_op_val[position])
            true_b_val_list.append(true_boundary_val)
        b_val=torch.cat(b_val_list)
        true_b_val=torch.cat(true_b_val_list)
    return b_val,true_b_val


def matrix_loss(model, grid, operator, bconds, lambda_bound=10):
    if bconds == None:
        print('No bconds is not possible, returning infinite loss')
        return np.inf

    op = apply_matrix_diff_operator(model, grid, operator)

    # we apply no  boundary conditions operators if they are all None

    b_val, true_b_val = apply_matrix_bcond_operator(model, grid, bconds)
    """
    actually, we can use L2 norm for the operator and L1 for boundary
    since L2>L1 and thus, boundary values become not so signifnicant, 
    so the NN converges faster. On the other hand - boundary conditions is the
    crucial thing for all that stuff, so we should increase significance of the
    coundary conditions
    """
    # l1_lambda = 0.001
    # l1_norm =sum(p.abs().sum() for p in model.parameters())
    # loss = torch.mean((op) ** 2) + lambda_bound * torch.mean((b_val - true_b_val) ** 2)+ l1_lambda * l1_norm

    loss = torch.mean((op) ** 2) + lambda_bound * torch.mean((b_val - true_b_val) ** 2)

    return loss



def nn_autograd_simple(model, points, order,axis=0):
    points.requires_grad=True
    f = model(points).sum()
    for i in range(order):
        grads, = torch.autograd.grad(f, points, create_graph=True)
        f = grads[:,axis].sum()
    return grads[:,axis]


def nn_autograd_mixed(model, points,axis=[0]):
    points.requires_grad=True
    f = model(points).sum()
    for ax in axis:
        grads, = torch.autograd.grad(f, points, create_graph=True)
        f = grads[:,ax].sum()
    return grads[:,axis[-1]]



def nn_autograd(*args,axis=0):
    model=args[0]
    points=args[1]
    if len(args)==3:
        order=args[2]
        grads=nn_autograd_simple(model, points, order,axis=axis)
    else:
        grads=nn_autograd_mixed(model, points,axis=axis)
    return grads.reshape(-1,1)


def take_derivative_autograd (model, grid, term):
    """
    Axiluary function serves for single differential operator resulting field
    derivation

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    term : TYPE
        differential operator in conventional form.

    Returns
    -------
    der_term : torch.Tensor
        resulting field, computed on a grid.

    """
    # it is may be int, function of grid or torch.Tensor
    coeff = term[0]
    # this one contains shifted grids (see input_preprocessing module)
    product = term[1]
    # float that represents power of the differential term
    power = term[2]
    # initially it is an ones field
    der_term = torch.zeros_like(model(grid)) + 1
    
    for j,derivative in enumerate(product):
        if derivative==[None]:
            der=model(grid)
        else:
            der=nn_autograd(model,grid,axis=derivative)
        der_term = der_term * der ** power[j]
    
    der_term = coeff * der_term
    
    
    return der_term


def apply_autograd_operator(model,grid, operator):
    """
    Deciphers equation in a single grid subset to a field.

    Parameters
    ----------
    model : torch.Sequential
        Neural network.
    operator : list
        Single (len(subset)==1) operator in input form. See 
        input_preprocessing.operator_prepare()

    Returns
    -------
    total : torch.Tensor

    """
    for term in operator:
        dif = take_derivative_autograd(model, grid, term)
        try:
            total += dif
        except NameError:
            total = dif
    return total


def apply_autograd_bcond_operator(model,grid,bconds):
    true_b_val_list = []
    b_val_list = []
    b_pos_list = []

    # we apply no  boundary conditions operators if they are all None

    simpleform = False
    for bcond in bconds:
        if bcond[1] == None:
            simpleform = True
        if bcond[1] != None:
            simpleform = False
            break
    if simpleform:
        for bcond in bconds:
            b_pos_list.append(bcond[0])
            true_boundary_val = bcond[2].reshape(-1, 1)
            true_b_val_list.append(true_boundary_val)
        # print(flatten_list(b_pos_list))
        # b_pos=torch.cat(b_pos_list)
        true_b_val = torch.cat(true_b_val_list)
        b_op_val = model(grid)
        b_val = b_op_val[flatten_list(b_pos_list)]
    # or apply differential operatorn first to compute corresponding field and
    else:
        for bcond in bconds:
            b_pos = bcond[0]
            b_pos_list.append(bcond[0])
            b_cond_operator = bcond[1]
            true_boundary_val = bcond[2].reshape(-1, 1)
            true_b_val_list.append(true_boundary_val)
            if b_cond_operator == None or b_cond_operator == [[1, [None], 1]]:
                b_op_val = model(grid)
            else:
                b_op_val =apply_autograd_operator(model,grid,b_cond_operator)
            # take boundary values
            b_val_list.append(b_op_val[b_pos])
        true_b_val = torch.cat(true_b_val_list)
        b_val = torch.cat(b_val_list)
    return b_val,true_b_val



def autograd_loss(model, grid, operator, bconds, lambda_bound=10):
    if bconds == None:
        print('No bconds is not possible, returning infinite loss')
        return np.inf

    op = apply_autograd_operator(model,grid, operator)

    # we apply no  boundary conditions operators if they are all None

    b_val, true_b_val = apply_autograd_bcond_operator(model,grid,bconds)
    """
    actually, we can use L2 norm for the operator and L1 for boundary
    since L2>L1 and thus, boundary values become not so signifnicant, 
    so the NN converges faster. On the other hand - boundary conditions is the
    crucial thing for all that stuff, so we should increase significance of the
    coundary conditions
    """
    # l1_lambda = 0.001
    # l1_norm =sum(p.abs().sum() for p in model.parameters())
    # loss = torch.mean((op) ** 2) + lambda_bound * torch.mean((b_val - true_b_val) ** 2)+ l1_lambda * l1_norm
    bcond_part=b_val - true_b_val
    loss = torch.mean((op) ** 2) + lambda_bound * torch.mean((bcond_part) ** 2)

    return loss






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 26 13:36:47 2021

@author: mike_ubuntu
"""

from setuptools import setup, find_packages
import os
from pathlib import Path
import pathlib

HERE = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
SHORT_DESCRIPTION = 'TEDEouS - Torch Exhaustive Differential Equations Solver. Differential equation solver, based on pytorch library'
LONG_DESCRIPTION = (Path(os.path.join(HERE, "README.rst"))).read_text()

name = 'TEDEouS' # Add name
version= '0.0.15' # Add version
description = 'TEDEouS - Torch Exhaustive Differential Equations Solver. Differential equation solver, based on pytorch library'
author = 'Alexander Hvatov'
author_email = 'itmo.nss.team@gmail.com' # add email
url = 'https://github.com/ITMO-NSS-team/torch_DE_solver'

def read(*names, **kwargs):
    with open(
            os.path.join(os.path.dirname(__file__), *names),
            encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


def extract_requirements(file_name):
    return [r for r in read(file_name).split('\n') if r and not r.startswith('#')]


def get_requirements():
    requirements = extract_requirements('requirements.txt')
    return requirements

setup(
      name = name,
      version = version,
      author = author,
      author_email = author_email,
      url = url,
      description = SHORT_DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/x-rst",
      classifiers = [      
              'Development Status :: 3 - Alpha',
              'Programming Language :: Python :: 3',
              'License :: OSI Approved :: MIT License',
              'Operating System :: OS Independent',
      ],
      packages = find_packages(include = ['TEDEouS']), 
      python_requires =' >=3.8'
      )
